@@ -4,9 +4,11 @@ import json
 import os
 import re
 import shutil
+import signal
 import sys
 from enum import Enum
 from pathlib import Path
+from typing import Union
 
 import click
 
@@ -34,7 +36,15 @@ def _meson_cli():
         return [meson_cli]
 
 
-def _editable_install_path(distname):
+def editable_install_path(distname: str) -> Union[str, None]:
+    """Return path of the editable install for package `distname`.
+
+    If the package is not an editable install, return None.
+
+    See Also
+    --------
+    is_editable_install
+    """
     import importlib_metadata
 
     try:
@@ -56,16 +66,34 @@ def _editable_install_path(distname):
         return None
 
 
-def _is_editable_install(distname):
-    return _editable_install_path(distname) is not None
+# backward compat
+_editable_install_path = editable_install_path
 
 
-def _is_editable_install_of_same_source(distname):
-    editable_path = _editable_install_path(distname)
-    return editable_path and os.path.samefile(_editable_install_path(distname), ".")
+def is_editable_install(distname: str, verify_path: bool = False) -> bool:
+    """Whether or not an editable install of `distname` is present.
+
+    Parameters
+    ----------
+    `distname` : str
+        Name of the package. E.g., ``numpy`` or ``scikit-image``.
+        Not always the same as the module name (``numpy`` and
+        ``skimage`` for the above).
+    """
+    return editable_install_path(distname) is not None
 
 
-def _set_pythonpath(build_dir, quiet=False):
+# backward compat
+_is_editable_install = is_editable_install
+
+
+def _is_editable_install_of_same_source(distname: str) -> bool:
+    """Check whether the editable install was made from the current directory."""
+    editable_path = editable_install_path(distname)
+    return (editable_path is not None) and os.path.samefile(editable_path, ".")
+
+
+def _set_pythonpath(build_dir: str, quiet: bool = False) -> str:
     """Set first entry of PYTHONPATH to site packages directory.
 
     For editable installs, leave the PYTHONPATH alone.
@@ -77,12 +105,12 @@ def _set_pythonpath(build_dir, quiet=False):
     cfg = get_config()
     distname = cfg.get("project.name", None)
     if distname:
-        if _is_editable_install(distname):
+        if is_editable_install(distname):
             if _is_editable_install_of_same_source(distname):
                 if not (quiet):
                     click.secho(
                         "Editable install of same source directory detected; not setting PYTHONPATH",
-                        fg="bright_red",
+                        fg="yellow",
                     )
                 return ""
             else:
@@ -113,11 +141,11 @@ def _set_pythonpath(build_dir, quiet=False):
     return site_packages
 
 
-def _get_install_dir(build_dir):
+def _get_install_dir(build_dir: str) -> str:
     return f"{build_dir}-install"
 
 
-def _get_site_packages(build_dir):
+def _get_site_packages(build_dir: str) -> str:
     install_dir = _get_install_dir(build_dir)
     try:
         cfg = get_config()
@@ -139,13 +167,12 @@ def _get_site_packages(build_dir):
     site_packages = None
     if any(f"python{X}." in p for p in candidate_paths):
         # We have a system that uses `python3.X/site-packages` or `python3.X/dist-packages`
-        site_packages = [p for p in candidate_paths if f"python{X}.{Y}" in p]
-        if len(site_packages) == 0:
+        site_packages_paths = [p for p in candidate_paths if f"python{X}.{Y}" in p]
+        if len(site_packages_paths) == 0:
             raise FileNotFoundError(
                 f"No site-packages found in {install_dir} for Python {X}.{Y}"
             )
-        else:
-            site_packages = site_packages[0]
+        site_packages = site_packages_paths[0]
     else:
         # A naming scheme that does not encode the Python major/minor version is used, so return
         # whatever site-packages path was found
@@ -164,22 +191,22 @@ def _get_site_packages(build_dir):
     return site_packages
 
 
-def _meson_version():
+def _meson_version() -> Union[str, None]:
     try:
         p = _run(_meson_cli() + ["--version"], output=False, echo=False)
         return p.stdout.decode("ascii").strip()
     except:
-        pass
+        return None
 
 
-def _meson_version_configured(build_dir):
+def _meson_version_configured(build_dir: str) -> Union[str, None]:
     try:
         meson_info_fn = os.path.join(build_dir, "meson-info", "meson-info.json")
         with open(meson_info_fn) as f:
             meson_info = json.load(f)
         return meson_info["meson_version"]["full"]
     except:
-        pass
+        return None
 
 
 def _meson_coverage_configured() -> bool:
@@ -198,7 +225,7 @@ def _meson_coverage_configured() -> bool:
     return False
 
 
-def _check_coverage_tool_installation(coverage_type: GcovReportFormat, build_dir):
+def _check_coverage_tool_installation(coverage_type: GcovReportFormat, build_dir: str):
     requirements = {  # https://github.com/mesonbuild/meson/blob/6e381714c7cb15877e2bcaa304b93c212252ada3/docs/markdown/Unit-tests.md?plain=1#L49-L62
         GcovReportFormat.html: ["Gcovr/GenHTML", "lcov"],
         GcovReportFormat.xml: ["Gcovr (version 3.3 or higher)"],
@@ -230,18 +257,40 @@ def _check_coverage_tool_installation(coverage_type: GcovReportFormat, build_dir
         )
 
 
+if sys.platform.startswith("win"):
+    DEFAULT_PREFIX = "C:/"
+else:
+    DEFAULT_PREFIX = "/usr"
+
+
+build_option = click.option(
+    "--no-build",
+    "build",
+    is_flag=True,
+    callback=lambda ctx, param, value: not value,  # store opposite value in `build` var
+    default=False,
+    help="Disable building before executing command",
+)
+
 build_dir_option = click.option(
     "-C",
     "--build-dir",
     default="build",
     show_envvar=True,
+    metavar="BUILD_DIR",
     envvar="SPIN_BUILD_DIR",
     help="Meson build directory; package is installed into './{build-dir}-install'.",
 )
 
 
 @click.command()
-@click.option("-j", "--jobs", help="Number of parallel tasks to launch", type=int)
+@click.option(
+    "-j",
+    "--jobs",
+    metavar="N_JOBS",
+    help="Number of parallel tasks to launch",
+    type=int,
+)
 @click.option("--clean", is_flag=True, help="Clean build directory before build")
 @click.option(
     "-v", "--verbose", is_flag=True, help="Print detailed build and installation output"
@@ -251,6 +300,13 @@ build_dir_option = click.option(
     is_flag=True,
     help="Enable C code coverage using `gcov`. Use `spin test --gcov` to generate reports.",
 )
+@click.option(
+    "--prefix",
+    help="The build prefix, passed directly to meson.",
+    type=str,
+    metavar="PREFIX",
+    default=DEFAULT_PREFIX,
+)
 @click.argument("meson_args", nargs=-1)
 @build_dir_option
 def build(
@@ -262,13 +318,16 @@ def build(
     gcov=False,
     quiet=False,
     build_dir=None,
+    prefix=None,
+    meson_compile_args=(),
+    meson_install_args=(),
 ):
     """🔧 Build package with Meson/ninja
 
     The package is installed to `build-install` (unless a different
     build directory is specified with `-C`).
 
-    MESON_ARGS are passed through e.g.:
+    MESON_ARGS are passed through to `meson setup` e.g.:
 
       spin build -- -Dpkg_config_path=/lib64/pkgconfig
 
@@ -303,16 +362,18 @@ def build(
         if not quiet:
             click.secho(
                 "Editable install of same source detected; skipping build",
-                fg="bright_red",
+                fg="yellow",
             )
         return
 
-    meson_args = list(meson_args)
+    meson_args_setup = list(meson_args)
 
     if gcov:
-        meson_args = meson_args + ["-Db_coverage=true"]
+        meson_args_setup = meson_args_setup + ["-Db_coverage=true"]
 
-    setup_cmd = _meson_cli() + ["setup", build_dir, "--prefix=/usr"] + meson_args
+    setup_cmd = (
+        _meson_cli() + ["setup", build_dir, f"--prefix={prefix}"] + meson_args_setup
+    )
 
     if clean:
         print(f"Removing `{build_dir}`")
@@ -344,10 +405,15 @@ def build(
         compile_flags += ["-j", str(jobs)]
 
     p = _run(
-        _meson_cli() + ["compile"] + compile_flags + ["-C", build_dir],
+        _meson_cli()
+        + ["compile"]
+        + compile_flags
+        + ["-C", build_dir]
+        + list(meson_compile_args),
         sys_exit=True,
         output=not quiet,
     )
+
     p = _run(
         _meson_cli()
         + [
@@ -359,7 +425,8 @@ def build(
             install_dir
             if os.path.isabs(install_dir)
             else os.path.relpath(abs_install_dir, abs_build_dir),
-        ],
+        ]
+        + list(meson_install_args),
         output=(not quiet) and verbose,
     )
 
@@ -416,6 +483,7 @@ Which tests to run. Can be a module, function, class, or method:
     default="html",
     help=f"Format of the gcov report. Can be one of {', '.join(e.value for e in GcovReportFormat)}.",
 )
+@build_option
 @build_dir_option
 @click.pass_context
 def test(
@@ -428,6 +496,7 @@ def test(
     coverage=False,
     gcov=None,
     gcov_format=None,
+    build=None,
     build_dir=None,
 ):
     """🔧 Run tests
@@ -444,7 +513,7 @@ def test(
 
     To run test modules, functions, classes, or methods:
 
-      spin -t numpy.random
+      spin test -t numpy.random
 
     To report the durations of the N slowest tests:
 
@@ -480,18 +549,23 @@ def test(
 
     package = cfg.get("tool.spin.package", None)
     if package is None:
-        print(
-            "Please specify `package = packagename` under `tool.spin` section of `pyproject.toml`"
+        click.secho(
+            "Please specify `package = packagename` under `tool.spin` section of `pyproject.toml`",
+            fg="bright_red",
         )
         raise SystemExit(1)
 
     # User did not specify what to test, so we test
-    # the full package
+    # the full package, or the tests directory if that is present
     if not (pytest_args or tests):
-        pytest_args = ("--pyargs", package)
+        if os.path.isdir("./tests"):
+            # tests dir exists, presuming you are not shipping tests
+            # with your package, and prefer to run those instead
+            pytest_args = (os.path.abspath("./tests"),)
+        else:
+            pytest_args = ("--pyargs", package)
     elif tests:
         if (os.path.sep in tests) or ("/" in tests):
-            # Tests specified as file path
             pytest_args = pytest_args + (tests,)
         else:
             # Otherwise tests given as modules
@@ -503,17 +577,18 @@ def test(
             "Error: cannot generate coverage report for editable installs",
             fg="bright_red",
         )
-        raise SystemExit(-1)
+        raise SystemExit(1)
 
-    build_cmd = _get_configured_command("build")
-    if build_cmd:
-        click.secho(
-            "Invoking `build` prior to running tests:", bold=True, fg="bright_green"
-        )
-        if gcov is not None:
-            ctx.invoke(build_cmd, build_dir=build_dir, gcov=bool(gcov))
-        else:
-            ctx.invoke(build_cmd, build_dir=build_dir)
+    if build:
+        build_cmd = _get_configured_command("build")
+        if build_cmd:
+            click.secho(
+                "Invoking `build` prior to running tests:", bold=True, fg="bright_green"
+            )
+            if gcov is not None:
+                ctx.invoke(build_cmd, build_dir=build_dir, gcov=bool(gcov))
+            else:
+                ctx.invoke(build_cmd, build_dir=build_dir)
 
     site_path = _set_pythonpath(build_dir)
 
@@ -536,9 +611,6 @@ def test(
     if (n_jobs != "1") and ("-n" not in pytest_args):
         pytest_args = ("-n", str(n_jobs)) + pytest_args
 
-    if not any("--import-mode" in arg for arg in pytest_args):
-        pytest_args = ("--import-mode=importlib",) + pytest_args
-
     if verbose:
         pytest_args = ("-v",) + pytest_args
 
@@ -564,8 +636,11 @@ def test(
     if not os.path.exists(install_dir):
         os.mkdir(install_dir)
 
+    # Unless we have a src layout, we need to switch away from the current directory into build install to avoid importing ./package instead of the built package.
+    test_path = site_path if not os.path.isdir("./src") else None
+
     cwd = os.getcwd()
-    pytest_p = _run(cmd + list(pytest_args), cwd=site_path)
+    pytest_p = _run(cmd + list(pytest_args), cwd=test_path)
     os.chdir(cwd)
 
     if gcov:
@@ -607,11 +682,14 @@ def test(
 
 
 @click.command()
-@click.option("--code", "-c", help="Python program passed in as a string")
+@click.option(
+    "--code", "-c", metavar="CODE", help="Python program passed in as a string"
+)
 @click.argument("gdb_args", nargs=-1)
+@build_option
 @build_dir_option
 @click.pass_context
-def gdb(ctx, *, code, gdb_args, build_dir):
+def gdb(ctx, *, code, gdb_args, build=None, build_dir=None):
     """👾 Execute code through GDB
 
       spin gdb -c 'import numpy as np; print(np.__version__)'
@@ -632,12 +710,13 @@ def gdb(ctx, *, code, gdb_args, build_dir):
      spin gdb my_tests.py
      spin gdb -- my_tests.py --mytest-flag
     """
-    build_cmd = _get_configured_command("build")
-    if build_cmd:
-        click.secho(
-            "Invoking `build` prior to invoking gdb:", bold=True, fg="bright_green"
-        )
-        ctx.invoke(build_cmd, build_dir=build_dir)
+    if build:
+        build_cmd = _get_configured_command("build")
+        if build_cmd:
+            click.secho(
+                "Invoking `build` prior to invoking gdb:", bold=True, fg="bright_green"
+            )
+            ctx.invoke(build_cmd, build_dir=build_dir)
 
     _set_pythonpath(build_dir)
     gdb_args = list(gdb_args)
@@ -662,33 +741,40 @@ def gdb(ctx, *, code, gdb_args, build_dir):
 
 @click.command()
 @click.argument("ipython_args", nargs=-1)
+@build_option
 @build_dir_option
 @click.pass_context
-def ipython(ctx, *, ipython_args, build_dir):
+def ipython(ctx, *, ipython_args, build=None, build_dir=None, pre_import=""):
     """💻 Launch IPython shell with PYTHONPATH set
 
     IPYTHON_ARGS are passed through directly to IPython, e.g.:
 
     spin ipython -- -i myscript.py
     """
-    build_cmd = _get_configured_command("build")
-    if build_cmd:
-        click.secho(
-            "Invoking `build` prior to invoking ipython:", bold=True, fg="bright_green"
-        )
-        ctx.invoke(build_cmd, build_dir=build_dir)
+    if build:
+        build_cmd = _get_configured_command("build")
+        if build_cmd:
+            click.secho(
+                "Invoking `build` prior to launching ipython:",
+                bold=True,
+                fg="bright_green",
+            )
+            ctx.invoke(build_cmd, build_dir=build_dir)
 
     p = _set_pythonpath(build_dir)
     if p:
         print(f'💻 Launching IPython with PYTHONPATH="{p}"')
+    if pre_import:
+        ipython_args = (f"--TerminalIPythonApp.exec_lines={pre_import}",) + ipython_args
     _run(["ipython", "--ignore-cwd"] + list(ipython_args), replace=True)
 
 
 @click.command()
 @click.argument("shell_args", nargs=-1)
+@build_option
 @build_dir_option
 @click.pass_context
-def shell(ctx, shell_args=[], build_dir=None):
+def shell(ctx, shell_args=[], build=None, build_dir=None):
     """💻 Launch shell with PYTHONPATH set
 
     SHELL_ARGS are passed through directly to the shell, e.g.:
@@ -698,12 +784,15 @@ def shell(ctx, shell_args=[], build_dir=None):
     Ensure that your shell init file (e.g., ~/.zshrc) does not override
     the PYTHONPATH.
     """
-    build_cmd = _get_configured_command("build")
-    if build_cmd:
-        click.secho(
-            "Invoking `build` prior to invoking shell:", bold=True, fg="bright_green"
-        )
-        ctx.invoke(build_cmd, build_dir=build_dir)
+    if build:
+        build_cmd = _get_configured_command("build")
+        if build_cmd:
+            click.secho(
+                "Invoking `build` prior to invoking shell:",
+                bold=True,
+                fg="bright_green",
+            )
+            ctx.invoke(build_cmd, build_dir=build_dir)
 
     p = _set_pythonpath(build_dir)
     if p:
@@ -718,21 +807,25 @@ def shell(ctx, shell_args=[], build_dir=None):
 
 @click.command()
 @click.argument("python_args", nargs=-1)
+@build_option
 @build_dir_option
 @click.pass_context
-def python(ctx, *, python_args, build_dir):
+def python(ctx, *, python_args, build=None, build_dir=None):
     """🐍 Launch Python shell with PYTHONPATH set
 
     PYTHON_ARGS are passed through directly to Python, e.g.:
 
     spin python -- -c 'import sys; print(sys.path)'
     """
-    build_cmd = _get_configured_command("build")
-    if build_cmd:
-        click.secho(
-            "Invoking `build` prior to invoking Python:", bold=True, fg="bright_green"
-        )
-        ctx.invoke(build_cmd, build_dir=build_dir)
+    if build:
+        build_cmd = _get_configured_command("build")
+        if build_cmd:
+            click.secho(
+                "Invoking `build` prior to invoking Python:",
+                bold=True,
+                fg="bright_green",
+            )
+            ctx.invoke(build_cmd, build_dir=build_dir)
 
     p = _set_pythonpath(build_dir)
     if p:
@@ -759,10 +852,11 @@ def python(ctx, *, python_args, build_dir):
 
 
 @click.command(context_settings={"ignore_unknown_options": True})
+@build_option
 @build_dir_option
 @click.argument("args", nargs=-1)
 @click.pass_context
-def run(ctx, *, args, build_dir=None):
+def run(ctx, *, args, build=None, build_dir=None):
     """🏁 Run a shell command with PYTHONPATH set
 
     \b
@@ -781,12 +875,13 @@ def run(ctx, *, args, build_dir=None):
     if not len(args) > 0:
         raise RuntimeError("No command given")
 
-    build_cmd = _get_configured_command("build")
-    if build_cmd:
-        # Redirect spin generated output
-        with contextlib.redirect_stdout(sys.stderr):
-            # Also ask build to be quiet
-            ctx.invoke(build_cmd, build_dir=build_dir, quiet=True)
+    if build:
+        build_cmd = _get_configured_command("build")
+        if build_cmd:
+            # Redirect spin generated output
+            with contextlib.redirect_stdout(sys.stderr):
+                # Also ask build to be quiet
+                ctx.invoke(build_cmd, build_dir=build_dir, quiet=True)
 
     is_posix = sys.platform in ("linux", "darwin")
     shell = len(args) == 1
@@ -798,7 +893,27 @@ def run(ctx, *, args, build_dir=None):
             cmd_args = ["bash", "-c", cmd_args]
 
     _set_pythonpath(build_dir, quiet=True)
-    p = _run(cmd_args, echo=False, shell=shell, sys_exit=False)
+
+    is_windows = sys.platform == "win32"
+
+    if not is_windows:
+        # Let the subprocess handle its own signals
+        # Except on Windows, where it already seems to work as intended,
+        # and `preexec_fn` is not supported
+        signal.signal(signal.SIGINT, signal.SIG_IGN)
+
+    def attach_sigint():
+        # Reset SIGINT handler to default
+        signal.signal(signal.SIGINT, signal.SIG_DFL)
+
+    # --- launch subprocess ---
+    p = _run(
+        cmd_args,
+        echo=False,
+        shell=shell,
+        sys_exit=False,
+        preexec_fn=None if is_windows else attach_sigint,
+    )
 
     # Is the user trying to run a Python script, without calling the Python interpreter?
     executable = args[0]
@@ -822,19 +937,20 @@ def run(ctx, *, args, build_dir=None):
     default=False,
     help="Clean previously built docs before building",
 )
-@click.option(
-    "--build/--no-build",
-    "first_build",
-    default=True,
-    help="Build project before generating docs",
-)
 @click.option(
     "--plot/--no-plot",
     "sphinx_gallery_plot",
     default=True,
     help="Sphinx gallery: enable/disable plots",
 )
-@click.option("--jobs", "-j", default="auto", help="Number of parallel build jobs")
+@click.option(
+    "--jobs",
+    "-j",
+    default="auto",
+    metavar="N_JOBS",
+    help="Number of parallel build jobs",
+)
+@build_option
 @build_dir_option
 @click.pass_context
 def docs(
@@ -842,10 +958,10 @@ def docs(
     *,
     sphinx_target,
     clean,
-    first_build,
     jobs,
     sphinx_gallery_plot,
     clean_dirs=None,
+    build=None,
     build_dir=None,
 ):
     """📖 Build Sphinx documentation
@@ -875,7 +991,7 @@ def docs(
 
     if sphinx_target in ("targets", "help"):
         clean = False
-        first_build = False
+        build = False
         sphinx_target = "help"
 
     if clean:
@@ -897,7 +1013,7 @@ def docs(
 
     build_cmd = _get_configured_command("build")
 
-    if build_cmd and first_build:
+    if build_cmd and build:
         click.secho(
             "Invoking `build` prior to building docs:", bold=True, fg="bright_green"
         )
@@ -923,7 +1039,7 @@ def docs(
 
     if site_path:
         os.environ["PYTHONPATH"] = (
-            f'{site_path}{os.sep}:{os.environ.get("PYTHONPATH", "")}'
+            f"{site_path}{os.sep}:{os.environ.get('PYTHONPATH', '')}"
         )
         click.secho(
             f"$ export PYTHONPATH={os.environ['PYTHONPATH']}",
@@ -937,11 +1053,14 @@ def docs(
 
 
 @click.command()
-@click.option("--code", "-c", help="Python program passed in as a string")
+@click.option(
+    "--code", "-c", metavar="CODE", help="Python program passed in as a string"
+)
 @click.argument("lldb_args", nargs=-1)
+@build_option
 @build_dir_option
 @click.pass_context
-def lldb(ctx, *, code, lldb_args, build_dir=None):
+def lldb(ctx, *, code, lldb_args, build=None, build_dir=None):
     """👾 Execute code through LLDB
 
       spin lldb -c 'import numpy as np; print(np.__version__)'
@@ -964,12 +1083,13 @@ def lldb(ctx, *, code, lldb_args, build_dir=None):
      spin lldb -- --arch x86_64 -- my_tests.py
      spin lldb -c 'import numpy as np; print(np.__version__)' -- --arch x86_64
     """
-    build_cmd = _get_configured_command("build")
-    if build_cmd:
-        click.secho(
-            "Invoking `build` prior to invoking lldb:", bold=True, fg="bright_green"
-        )
-        ctx.invoke(build_cmd, build_dir=build_dir)
+    if build:
+        build_cmd = _get_configured_command("build")
+        if build_cmd:
+            click.secho(
+                "Invoking `build` prior to invoking lldb:", bold=True, fg="bright_green"
+            )
+            ctx.invoke(build_cmd, build_dir=build_dir)
 
     _set_pythonpath(build_dir)
     lldb_args = list(lldb_args)
